#ifndef ARRAY_H_
#define ARRAY_H_

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <string>
#include <iostream>
#include <iomanip>
#include <complex>  

template <class T>
class Array
{
        
private :
	size_t  __rows;
	size_t  __cols;
	T*      __elts;
        
public :

    Array(size_t rows=0, size_t cols=0, T val=T(0))
    {
        __rows = rows;
        __cols = cols;
        __elts = new T [__rows * __cols];
        for (T* pt = __elts; pt < __elts + __rows * __cols; *pt++ = val);
    };
    
    Array(const Array<T>& array)
    {
        __rows = array.__rows;
        __cols = array.__cols;
        __elts = new T [__rows * __cols];
        memcpy(__elts, array.__elts, __rows * __cols * sizeof(T));
    };
    
    ~Array()
    {
        delete [] __elts;
    };

    inline const size_t rows() const
    {
        return __rows;
    };
    
    inline const size_t cols() const
    {
        return __cols;
    };
        
    inline Array<T>& operator=(const Array<T>& array)
    {
        if (this != &array)
            delete [] __elts;
        __rows = array.__rows;
        __cols = array.__cols;
        __elts = new T [__rows * __cols];
        memcpy(__elts, array.__elts, __rows * __cols * sizeof(T));
        
        return (*this);
    };
    
    inline T& operator()(size_t row, size_t col)
    {
        return __elts[row * __cols + col];
    };
    
    inline T operator ()(int row, int col) const
    {
        return __elts[row * __cols + col];
    };
    
    inline Array<T> &operator+=(const Array<T>& array)
    {
        T* p1;
        T* p2;
        for (p1 = __elts, p2 = array.__elts; 
             p1 < __elts + __rows * __cols;
             *p1++ += *p2++);
        
        return (*this);
    };
    
    inline Array<T> &operator-=(const Array<T>& array)
    {
        T* p1;
        T* p2;
        for (p1 = __elts, p2 = array.__elts; 
             p1 < __elts + __rows * __cols;
             *p1++ -= *p2++);
        
        return (*this);
    };
    
    inline Array<T> &operator*=(const Array<T>& array)
    {
        T* p1;
        T* p2;
        for (p1 = __elts, p2 = array.__elts; 
             p1 < __elts + __rows * __cols;
             *p1++ *= *p2++);
        
        return (*this);
    };
    
    template<class U>
    inline Array<T> &operator*=(const Array<U>& array)
    {
        for (size_t r = 0; r < __rows; r++)
        {
            for (size_t c = 0; c < __cols; c++)
            {
                (*this)(r, c) *= array(r, c);
            }
        }

        return (*this);
    };
    
    inline Array<T> &operator/=(const Array<T>& array)
    {
        T* p1;
        T* p2;
        for (p1 = __elts, p2 = array.__elts; 
             p1 < __elts + __rows * __cols;
             *p1++ /= *p2++);
        
        return (*this);
    };
        
    inline Array<T>& operator*=(const T& t)
    {
        for (T* p = __elts; p < __elts + __rows * __cols; *p++ *= t);
        
        return (*this);
    };
        
    inline Array<T>& operator/=(const T& t)
    {
        for (T* p = __elts; p < __elts + __rows * __cols; *p++ /= t);
        
        return (*this);
    };
    
    inline Array<T>& flipud()
    {        
        Array<T> tmp(__rows, __cols);
        
        for (size_t r = 0; r < __rows; r++)
            for (size_t c = 0; c < __cols; c++)
                tmp(r, c) = this->operator()(__rows - r  - 1, c);
        
        *this = tmp;
        
        return (*this);
    };
    
    inline Array<T>& fliplr()
    {        
        Array<T> tmp(__rows, __cols);
        
        for (size_t r = 0; r < __rows; r++)
            for (size_t c = 0; c < __cols; c++)
                tmp(r, c) = this->operator()(r, __cols - c - 1);
        
        *this = tmp;
        
        return (*this);
    };

    inline T max() const
    {
        T max = __elts[0];
        for (T* p = __elts; p < __elts + __rows * __cols; p++)
            if (*p > max)
                max = *p;
        
        return (max);
    }
    
    inline T min() const
    {
        T min = __elts[0];
        for (T* p = __elts; p < __elts + __rows * __cols; p++)
            if (*p < min)
                min = *p;
        
        return (min);
    }
    
    inline T sum() const
    {
        T sum = 0;
        for (T* p = __elts; p < __elts + __rows * __cols; p++)
            sum += *p;
        
        return (sum);
    }
    
    inline T mean() const
    {
        return (sum() / (__rows * __cols));
    }
    
    inline Array<T>& fftshift()
    {
        size_t hrows = __rows / 2;
        size_t hcols = __cols / 2;
        Array<T> tmp(__rows, __cols);
        
        for (size_t r = 0; r < hrows; r++)
            for (size_t c = 0; c < hcols; c ++)
                tmp(r + hrows, c + hcols) = this->operator()(r, c);
        for (size_t r = 0; r < hrows; r++)
            for (size_t c = hcols; c < __cols; c ++)
                tmp(r + hrows, c - hcols) = this->operator()(r, c);
        for (size_t r = hrows; r < __rows; r++)
            for (size_t c = 0; c < hcols; c ++)
                tmp(r - hrows, c + hcols) = this->operator()(r, c);
        for (size_t r = hrows; r < __rows; r++)
            for (size_t c = hcols; c < __cols; c ++)
                tmp(r - hrows, c - hcols) = this->operator()(r, c);
        
        *this = tmp;
        
        return *this;
    }
    
    template<class C>
    friend std::ostream& operator<<(std::ostream& os, const Array<C> &array);

};

template <class T>
inline Array<T> operator +(const Array<T>& array1, const Array<T>& array2)
{
	Array<T> temp = array1;
	temp += array2;
    
	return temp;
}

template <class T>
inline Array<T> operator -(const Array<T>& array1, const Array<T>& array2)
{
	Array<T> temp = array1;
	temp -= array2;
    
	return temp;
}

template <class T>
inline Array<T> operator *(const Array<T>& array1, const Array<T>& array2)
{
	Array<T> temp = array1;
	temp *= array2;
    
	return temp;
}

template <class T, class U>
inline Array<T> operator *(const Array<T>& array1, const Array<U>& array2)
{
	Array<T> temp = array1;
	temp *= array2;
    
	return temp;
}

template <class T>
inline Array<T> operator /(const Array<T>& array1, const Array<T>& array2)
{
    Array<T> temp = array1;
    temp /= array2;
    
    return temp;
}

template <class T>
inline Array<T> operator *(const Array<T>& array, const T& t)
{
	Array<T> temp = array;
	temp *= t;
    
	return temp;
}

template <class T>
inline Array<T>operator *(const T& t, const Array<T>& array)
{
	return (array * t);
}

template <class T>
Array<T> operator /(const Array<T>& array, const T& t)
{
	Array<T> temp = array;
	temp /= t;
    
	return temp;
}

template <class T>
inline Array<T> operator /(const T& t, const Array<T>& array)
{
	return (array / t);
}

template <class C>
std::ostream&
operator<<(std::ostream& os, const Array<C> &array)
{
    os << std::fixed;
    for (size_t r = 0; r < array.__rows; r++)
    {
        for (size_t c = 0; c < array.__cols; c++)
        {
            os << array(r, c) << "\t";
        }
        os << std::endl;
    }
    
    return os;
}

#endif
