#include <vector>
#include <cmath>

#include "CannyBuilder.h"

#define SCALE (float)(8.0 * sqrt(2))
#define PI acos(-1.0)

CannyBuilder::CannyBuilder(float sigma, float tlow, float thigh)
: __sigma(sigma), __tlow(tlow), __thigh(thigh)
{}

Array<int>
CannyBuilder::operator()(const Array<float>& in)
{
    size_t cols = in.cols();
    size_t rows = in.rows();
    size_t r;
    size_t c;
    size_t i;
    float ssq = __sigma * __sigma;

    //////////////////////////////////////////////////////////////////////////
    // Gaussian smoothing
    //////////////////////////////////////////////////////////////////////////
    
    // Decide the size of Gaussian template
    //Matlab
    /* 
    float gaussian_die_off = 0.0001;
    size_t width = 1;
    while(exp(-(float)(width  * width)  / 
                    (2.0 * ssq)) > gaussian_die_off)
        width ++;
    width --;
    size_t sizegt = 2 * width + 1;
    */
    
    // Java
    size_t sizegt = (size_t) (SCALE * __sigma);
    if (sizegt % 2 == 0)
        sizegt ++;
    size_t width = (size_t) ((sizegt - 1) / 2);
    
    // Gaussian 1D filter.
    std::vector<float> gauss(sizegt);
    // Matlab
    /*
    for (i = 0; i < sizegt; i ++)
        gauss[i] = (float) (exp(-(float)(((int)i -  width) * ((int)i - width))  / (2.0 * ssq)) / 
                                     2.0 * PI * ssq);
    */
    // Java
    for (i=0; i < sizegt; i ++)
        gauss[i] = (float) (1.0 / (sqrtf(2.0 * PI) *__sigma) *
                            expf(powf(((int)i - width) / __sigma, 2) / (-2.0)));

    float min = 100000000;
    float max = -100000000;
   
    Array<float> temp(rows, cols);
    // Perform Gaussian smoothing to the input Builder horizontally
    for (r = 0; r < rows; r ++)
    {
        for (c = 0; c < cols; c ++)
        {
            for (i = 0; i < sizegt; i ++)
            {
                int tempc = c + i - (int)width;
                // if the index is not in the input Builder, reflex the Builder (virtually)
                if (tempc < 0) 
                    tempc = - tempc - 1;
                
                if (tempc > (int)cols - 1)
                    tempc = 2 * int(cols) - tempc - 1;
                
                temp(r, c) += gauss[i] * in(r, (size_t)tempc);
            }
        }
    }
    
    Array<float> smoothed(in);
    // Perform Gaussian smoothing to the input Builder vertically
    for (r = 0; r < rows; r ++)
    {
        for (c = 0; c < cols; c ++)
        {
            for (i = 0; i < sizegt; i ++) 
            {
                int tempr = r + i - width;
                // if the index is not in the input Builder,  reflex the Builder (virtually)
                if (tempr < 0)
                    tempr = - tempr - 1;
                
                if (tempr > (int)rows - 1 )
                    tempr = 2 * (int)rows - tempr - 1;
                
                smoothed(r, c) += gauss[i] * temp((size_t)tempr, c);
            }
            
            if (min > smoothed(r, c))
                min = smoothed(r, c);
            
            if (max < smoothed(r, c))
                max = smoothed(r, c);
        }
    }
    
    // scale the smoothed data
    for (r = 0; r < rows; r ++)
        for (c = 0; c < cols; c ++)
            smoothed(r, c) = (float)(255.0 / (max - min) * (smoothed(r, c) - min));
  
    //////////////////////////////////////////////////////////////////////////
    // Compute the gradient magnitude and direction by Roberts operators
    //////////////////////////////////////////////////////////////////////////
    Array<float> direction(rows, cols);
    
    float Ix = 0;                      // partial derivative along x-axis
    float Iy = 0;                      // partial derivative along y-axis
    float I = 0;                       // magnitude of the gradient    
    
    // Compute gradient magnitude and direction
    for (r = 0; r < rows; r ++)
    {
        for (c = 0; c < cols; c ++)
        {
            size_t tempc = c + 1;
            if (tempc > cols - 1)
                tempc = cols - 1;
            
            size_t tempr = r + 1;
            if (tempr > rows - 1)
                tempr = rows - 1;
            
            Ix = smoothed(r, tempc) - smoothed(tempr, c);
            
            Iy = smoothed(r, c) - smoothed(tempr, tempc);
            
            I = (float) sqrtf(Ix * Ix + Iy * Iy);
            
            smoothed(r, c) = I;
            
            // Compute gradient direction
            direction(r, c) = atan2f(Iy, Ix);
        }
    }

    //////////////////////////////////////////////////////////////////////////
    // Thin the edges
    //////////////////////////////////////////////////////////////////////////
    int delx[4] = {1,1,0,-1};
    int dely[4] = {0,1,1,1};
    float Imax = 0;
    int tempr, tempc;
        
    temp = smoothed;
    for (r = 0; r < rows; r ++)
    {
        for (c = 0; c < cols; c ++)
        {
            // gradient direction ranges from -Pi to -0
            float gd  = direction(r, c);
            if (gd < 0) gd += PI;
            
            // direction ranges from 0 to 3
            int dir = ((int)(gd / PI * 4.0 + 0.5)) % 4;
            
            tempr = r + delx[dir];
            if (tempr < 0)
                tempr = 0;
            if (tempr > (int)rows - 1)
                tempr = (int)rows - 1;
            
            tempc = c + dely[dir];
            if (tempc < 0)   
                tempc = 0;
            if (tempc > (int)cols - 1)
                tempc = (int)cols - 1;
            
            // suppress non-maxima
            if (temp(r, c) <= temp(size_t(tempr), size_t(tempc)))
                smoothed(r, c) = 0;

            tempr = r - delx[dir];
            if (tempr < 0)    
                tempr = 0;
            if (tempr > (int)rows - 1) 
                tempr = (int)rows - 1;
            
            tempc = c - dely[dir];
            if (tempc < 0)  
                tempc = 0;
            if (tempc > (int)cols - 1 )
                tempc = (int)cols - 1;
            
            // suppress non-maxima in the opposite direction
            if (temp(r, c) <= temp(size_t(tempr), size_t(tempc)))
                smoothed(r, c) = 0;
        }
    }
            
    // Find the maximum value of gradient magnitude
    for (r = 0; r < rows; r ++)
        for (c = 0; c < cols; c ++)
            if (Imax < smoothed(r, c))
                Imax = smoothed(r, c);
    
    // Scale magnitude to the range [0, 255]
    smoothed *= (float)(255.0 / Imax);
    
    //////////////////////////////////////////////////////////////////////////
    // Detect edges
    //////////////////////////////////////////////////////////////////////////
    Array<int> out(rows, cols);
        
    // perform edge detection
    for (r = 0; r < rows; r ++)
        for (c = 0; c < cols; c ++)
            if (out(r, c) == 0 && smoothed(r, c) > __thigh)
               __followEdge(out, smoothed, (int)r, (int)c);
    
    return out;
}

// A recursive function to find connected pixels to form an edge
void 
CannyBuilder::__followEdge(Array<int>&in, Array<float>& gradient, 
    int r, int c) 
{

    in((size_t)r, (size_t)c) = 255;

    // Check on 8 neighbors
    if (r > 0 && c > 0 && gradient((size_t)r - 1, (size_t)c - 1) > __tlow && in((size_t)r - 1, (size_t)c - 1) == 0) 
        __followEdge(in, gradient, r - 1, c - 1);
    
    if (r > 0 && gradient((size_t)r - 1, (size_t)c) > __tlow && in((size_t)r - 1, (size_t)c) == 0)
        __followEdge(in, gradient, r - 1, c);
    
    if (r > 0 && c < (int)in.cols() - 1 && gradient((size_t)r - 1, (size_t)c + 1) > __tlow && in((size_t)r - 1, (size_t)c + 1) == 0)
        __followEdge(in, gradient, r - 1, c + 1);
    
    if (c > 0 && gradient((size_t)r, (size_t)c - 1) > __tlow && in((size_t)r, (size_t)c - 1) == 0) 
        __followEdge(in, gradient, r, c - 1);
    
    if (c < (int)in.cols() - 1 && gradient((size_t)r, (size_t)c + 1) > __tlow && in((size_t)r, (size_t)c + 1) == 0) 
        __followEdge(in, gradient, r, c + 1);
    
    if (r < (int)in.rows() - 1 && c > 0 && gradient((size_t)r + 1, (size_t)c - 1) > __tlow && in((size_t)r + 1, (size_t)c - 1) == 0)
        __followEdge(in, gradient, r + 1, c - 1);
    
    if (r < (int)in.rows() - 1 && gradient((size_t)r + 1, (size_t)c) > __tlow && in((size_t)r + 1, (size_t)c) == 0)
        __followEdge(in, gradient, r + 1, c);
    
    if (r < (int)in.rows() - 1 && c < (int)in.cols() - 1 && gradient((size_t)r + 1, (size_t)c + 1) > __tlow && in((size_t)r + 1, (size_t)c + 1) == 0) 
        __followEdge(in, gradient, r + 1, c + 1);
}
